@@ -1,3 +1,4 @@
+import arrow
 import numpy as np
 from tqdm import trange, tqdm
 from abc import ABC, abstractmethod
@@ -41,9 +42,10 @@ def get_train_triplets_set(train_data):
     return set([(x[0], x[1], x[2]) for x in train_data])
 
 def pprint(meg, filename=None):
+    now = arrow.now.format('YYYY-MM-DD HH:mm:ss *==|@=====>  ')
     if filename is not None:
         with open(filename, 'a') as f:
-            f.write(meg+'\n')
+            f.write(now + meg+'\n')
     print(meg)
 
 
@@ -59,23 +61,22 @@ def print_result(ranks, des='describe', filename='conf.txt'):
     pprint(meg, filename)
 
 
-class Base(ABC):
+class BaseEval(ABC):
     __slots__ = ['num_ent', 'num_rel', 'train_data', 'negative_rate',
                  'test_data', 'valid_data', 'batch_size', 'global_triplets']
 
-    def __init__(self, num_ent=0, num_rel=0, train_data=None, test_data= None,
-                 valid_data=None, negative_rate=1, batch_size=0):
-        super(Base, self).__init__()
+    def __init__(self, num_ent=0, num_rel=0, data=None, negative_rate=1, batch_size=0):
+        super(BaseEval, self).__init__()
 
         self.num_rel = num_rel
         self.num_ent = num_ent
         self.batch_size = batch_size
         self.negative_rate = negative_rate
-        self.train_data = train_data
-        self.valid_data = valid_data
-        self.test_data = test_data
+        self.train_data = data['train']
+        self.valid_data = data['valid']
+        self.test_data = data['test']
 
-        self.global_triplets = get_train_triplets_set(train_data)
+        self.global_triplets = None
     
     @staticmethod
     def generate_graph(data):
@@ -91,80 +92,13 @@ class Base(ABC):
         return data_set, np.array([[x[0], x[1]] for x in pairs])
     
     @abstractmethod
-    def negative_smaple(self, *args, **warg):
+    def negative_sample(self, *args, **warg):
         pass
     
     @abstractmethod
-    def negative_smaple_iter(self, *args, **warg):
+    def sample_iter(self, *args, **warg):
         pass
 
-    @abstractmethod
-    def calculate(self, *args, **warg):
-        pass
-    
-    @abstractmethod
-    def evaluate(self, *args, **warg):
-        pass
-        
-
-class Triplet(Base):
-    '''triplets.
-    
-       Example:
-       
-            from kgraph.utils import Triplet
-            poss = Triplet(num_entities, num_relations, train_data,
-                           test_data, valid_data, batch_size=1000)
-            
-            model
-            
-            for batch_data, labels in poss.negative_sample_iter(batch_size=2000, negative_rate=1):
-                    pred = model(batch_data)
-                    loss(pred, labels)
-            
-            poss.evaluate(model.predict, test_data)
-       
-    '''
-    
-    def negative_smaple(self, pos_samples):
-        size = len(pos_samples)
-        num_to_generate = size * self.negative_rate
-        neg_samples = np.tile(pos_samples, (self.negative_rate, 1))
-        labels = np.ones(size * (self.negative_rate + 1),
-                         dtype=np.float32) * (-1.)
-        labels[: size] = 1
-        
-        values = np.random.randint(self.num_ent, size=num_to_generate)
-        choices = np.random.uniform(size=num_to_generate)
-        subj = choices > 0.5
-        obj = choices <= 0.5
-        neg_samples[subj, 0] = values[subj]
-        neg_samples[obj, 2] = values[obj]
-        
-        for i, p in enumerate(choices):
-            while True:
-                triplet = (neg_samples[i, 0], neg_samples[i, 1], neg_samples[i, 2])
-                if triplet not in self.global_triplets:
-                    break
-                if p > 0.5:
-                    neg_samples[i, 0] = np.random.choice(self.num_ent)
-                else:
-                    neg_samples[i, 2] = np.random.choice(self.num_ent)
-        return np.concatenate((pos_samples, neg_samples)), labels
-
-    def negative_smaple_iter(self, batch_size=0, negative_rate=1):
-        if batch_size == 0:
-            batch_size = self.batch_size
-        if negative_rate != 1:
-            self.negative_rate = negative_rate
-        
-        size = len(self.train_data)
-        num_batch = size // batch_size
-        train_data = np.random.permutation(self.train_data)
-        
-        for i in trange(num_batch):
-            yield self.negative_smaple(train_data[i * batch_size: (i + 1) * batch_size, :])
-
     def calculate(self, target_function, data, batch_size=512, pred_tail=True, device='cpu'):
         if not pred_tail:
             src, rel, dst = data.transpose(1, 0)
@@ -185,10 +119,14 @@ class Triplet(Base):
                 j = i * batch_size
                 batch_data = new_triplets[j: j + batch_size, :]
                 batch_data = torch.from_numpy(batch_data).to(device)
-                score = target_function(batch_data).squeeze()
-                if not isinstance(score, np.ndarray):
-                    score = score.cpu().numpy()
-                scores[j: j + batch_size] += score
+                with torch.no_grad():
+                    score = target_function(batch_data).squeeze()
+                    if not isinstance(score, np.ndarray):
+                        if score.is_cuda:
+                            score = score.cpu().numpy()
+                        else:
+                            score = score.numpy()
+                    scores[j: j + batch_size] += score
             
             rank, frank = rank_score(data[0][(pair[0], pair[1])], scores,
                                      num_entities=self.num_ent)
@@ -200,7 +138,7 @@ class Triplet(Base):
         del data
         return ranks, franks
     
-    def evaluate(self, target_function, test_data, batch_size=512, filename='conf.txt', device=None):
+    def evaluate(self, target_function, test_data=None, batch_size=512, filename='conf.txt', device=None):
         
         device = 'cuda' if torch.cuda.is_available() and device is None else "cpu"
         
@@ -234,7 +172,123 @@ class Triplet(Base):
         pprint('\t', filename)
         
 
-class Pair(Base):
+class TrainEval_By_Triplet(BaseEval):
+    '''triplets.
+    
+       Example:
+       
+            from kgraph.utils import Triplet
+            poss = Triplet(num_entities, num_relations, train_data,
+                           test_data, valid_data, batch_size=1000)
+            
+            model
+            
+            for batch_data, labels in poss.sample_iter(batch_size=2000, negative_rate=1):
+                    pred = model(batch_data)
+                    loss(pred, labels)
+            
+            poss.evaluate(model.predict, test_data)
+       
+    '''
+    
+    def negative_sample(self, pos_samples):
+        size = len(pos_samples)
+        num_to_generate = size * self.negative_rate
+        neg_samples = np.tile(pos_samples, (self.negative_rate, 1))
+        labels = np.ones(size * (self.negative_rate + 1),
+                         dtype=np.float32) * (-1.)
+        labels[: size] = 1
+        
+        values = np.random.randint(self.num_ent, size=num_to_generate)
+        choices = np.random.uniform(size=num_to_generate)
+        subj = choices > 0.5
+        obj = choices <= 0.5
+        neg_samples[subj, 0] = values[subj]
+        neg_samples[obj, 2] = values[obj]
+        
+        for i, p in enumerate(choices):
+            while True:
+                triplet = (neg_samples[i, 0], neg_samples[i, 1], neg_samples[i, 2])
+                if triplet not in self.global_triplets:
+                    break
+                if p > 0.5:
+                    neg_samples[i, 0] = np.random.choice(self.num_ent)
+                else:
+                    neg_samples[i, 2] = np.random.choice(self.num_ent)
+        return np.concatenate((pos_samples, neg_samples)), labels
+
+    def sample_iter(self, batch_size=0, negative_rate=1):
+        if batch_size == 0:
+            batch_size = self.batch_size
+        if negative_rate != 1:
+            self.negative_rate = negative_rate
+        
+        self.global_triplets = get_train_triplets_set(self.train_data)
+        
+        size = len(self.train_data)
+        num_batch = size // batch_size
+        train_data = np.random.permutation(self.train_data)
+        
+        for i in trange(num_batch):
+            yield self.negative_sample(train_data[i * batch_size: (i + 1) * batch_size, :])
+    
+
+class TrainEval_By_Pair(BaseEval):
     pass
 
 
+class TrainEval_For_Trans(BaseEval):
+    
+    def __init__(self, num_ent=0, num_rel=0, data=None, negative_rate=1, batch_size=0):
+        # super(TrainEval_For_Trans, self).__init__()
+        self.num_rel = num_rel
+        self.num_ent = num_ent
+        self.batch_size = batch_size
+        self.negative_rate = negative_rate
+        self.train_data = data['train']
+        self.valid_data = data['valid']
+        self.test_data = data['test']
+
+        self.global_triplets = get_train_triplets_set(self.train_data)
+    
+    def sample(self, mix=True):
+        
+        def filt(samples, all_train_triplet, for_neg_head=True):
+            for i, s in enumerate(samples):
+                triplet = (s[0], s[1], s[2])
+                while triplet in all_train_triplet:
+                    if for_neg_head:
+                        samples[i, 0] = np.random.choice(self.num_ent)
+                        triplet = ((samples[i, 0], s[1], s[2]))
+                    else:
+                        samples[i, 2] = np.random.choice(self.num_ent)
+                        triplet = (s[0], s[1], samples[i, 2])
+                    
+        
+        global_triplets = self.global_triplets
+        train_data = self.train_data
+        
+        num_to_generate =len(train_data)
+        values = np.random.randint(self.num_ent, size=num_to_generate)
+        choices = np.random.uniform(size=num_to_generate)
+        subj = choices > 0.5
+        obj = choices <= 0.5
+        
+        pos_head_samples = train_data[subj, :]
+        neg_head_samples = pos_head_samples.copy()
+        neg_head_samples[:, 0] = values[subj]
+        filt(neg_head_samples, global_triplets)
+        
+        pos_tail_samples = train_data[obj, :]
+        neg_tail_samples = pos_tail_samples.copy()
+        neg_tail_samples[:, 2] = values[obj]
+        filt(neg_tail_samples, global_triplets, for_neg_head=False)
+        
+        if mix:
+            pos_samples = np.concatenate((pos_head_samples, pos_tail_samples))
+            neg_samples = np.concatenate((neg_head_samples, neg_tail_samples))
+            ret_index = np.random.permutation(np.arange(num_to_generate))
+            
+            return pos_samples[ret_index], neg_samples[ret_index]
+        else:
+            return (pos_head_samples, neg_head_samples), (pos_tail_samples, neg_tail_samples)
